/*
 *	Armadillo-3x0 LED Driver
 *	Copyright (C) 2006,2008 Atmark Techno, Inc.
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <asm/io.h>

#include <asm/arch/armadillo3x0_misc.h>
#include <asm/arch/armadillo3x0_led.h>

#define AUTHOR      "Atmark Techno, Inc."
#define DESCRIPTION "Armadillo-3x0 LED driver"
#define LICENSE     "GPL v2"

#define LED_MINOR   (215)
#define LED_DEVICE  "led"

#define VERSION     DESCRIPTION ", (C) 2006 " AUTHOR

#define ON     (1)
#define OFF    (2)
#define STATUS (4)

static spinlock_t led_lock = SPIN_LOCK_UNLOCKED;
static unsigned long led_status = 0;

#define LED_IS_OPEN    (0x01)
#define LED_READ_COMP  (0x02)
#define LED_WRITE_COMP (0x04)

#define LED_ALL        (0xffffffff)

#undef DEBUG
#if defined(DEBUG)
#define DEBUG_FUNC()        printk(LED_DEVICE ": %s()\n", __FUNCTION__)
#define DEBUG_INFO(args...) printk(LED_DEVICE ": " args)
#else
#define DEBUG_FUNC()
#define DEBUG_INFO(args...)
#endif
#define PRINT_INFO(args...) printk(KERN_INFO LED_DEVICE ": " args)
#define PRINT_ERR(args...)  printk(KERN_ERR  LED_DEVICE ": " args)

struct led_info {
	unsigned long reg;
	unsigned long bit;
};

static struct led_info leds[] = {
  { A3X0_MISC_GCR, A3X0_MISC_GCR_LED },
};

static int 
led_set(unsigned long mask, unsigned long data)
{
  int i;
  unsigned char val;
  DEBUG_FUNC();

  for(i=0; i<ARRAY_SIZE(leds); i++){
    if(mask & (1<<i)){
      val = readb(leds[i].reg) & ~(leds[i].bit);
      val |= ((data & (1<<i)) ? leds[i].bit : 0);
      writeb(val, leds[i].reg);
    }
  }
  
  return 0;
}

static int
led_get(unsigned long mask, unsigned long *data)
{
  int i;
  unsigned char val;
  DEBUG_FUNC();

  if(data == 0)
    return -EINVAL;

  *data = 0;
  for(i=0;i<ARRAY_SIZE(leds); i++){
    if(mask & (1<<i)){
      val = readb(leds[i].reg) & (leds[i].bit);
      *data |= (val ? 1 : 0) << i;
    }
  }
      
  return 0;
}

static int led_open(struct inode *inode, struct file *file)
{
  DEBUG_FUNC();

	spin_lock_irq(&led_lock);

	if(led_status & LED_IS_OPEN){
		spin_unlock_irq(&led_lock);
		return -EBUSY;
	}

	led_status = LED_IS_OPEN;
	spin_unlock_irq(&led_lock);
	return 0;
}

static int led_release(struct inode *inode, struct file *file)
{
  DEBUG_FUNC();

	spin_lock_irq(&led_lock);
	led_status = 0;
	spin_unlock_irq(&led_lock);
	
	return 0;
}

static ssize_t led_read(struct file *file, char __user *buf,
                        size_t count, loff_t *ppos)
{
	unsigned long tmp;
	DEBUG_FUNC();

	if(count < 1) return -EINVAL;

	spin_lock_irq(&led_lock);
	if(led_status & LED_READ_COMP){
		spin_unlock_irq(&led_lock);
		return 0;
	}
	
	led_status |= LED_READ_COMP;

	led_get(LED_ALL, &tmp);
	
	tmp = sprintf(buf, "%lx", tmp);

	spin_unlock_irq(&led_lock);

	return tmp;
}

static ssize_t led_write(struct file *file, const char __user *buf,
			 size_t count, loff_t *ppos)
{
	unsigned long tmp;
	DEBUG_FUNC();

	if(count < 1) return -EINVAL;

	spin_lock_irq(&led_lock);
	if(led_status & LED_WRITE_COMP){
		spin_unlock_irq(&led_lock);
		return -EFBIG;
	}

	led_status |= LED_WRITE_COMP;
	
	tmp = simple_strtol(buf,NULL,16);

	led_set(LED_ALL, tmp);

	spin_unlock_irq(&led_lock);

	return count;
}

static int 
led_ioctl(struct inode *inode, struct file *file, 
	  unsigned int cmd, unsigned long arg)
{
  struct a3x0_led_param *param = (struct a3x0_led_param *)arg;

	DEBUG_FUNC();
	if(param == NULL)
	  return -EINVAL;

	switch(cmd){
	case A3X0_LED_SET:
	  led_set(0x01, param->buf);
	  break;
	case A3X0_LED_GET:
	  led_get(0x01, &param->buf);
	  break;
	default:
	  return -EINVAL;
	}

	return 0;
}

static struct file_operations led_fops = {
	.owner   = THIS_MODULE,
	.open    = led_open,
	.release = led_release,
	.read    = led_read,
	.write   = led_write,
	.ioctl   = led_ioctl,
};

static struct miscdevice led_dev = {
	.minor = LED_MINOR,
	.name  = LED_DEVICE,
	.fops  = &led_fops,
};

static int __init 
led_init(void)
{
	DEBUG_FUNC();
	if (!request_mem_region(A3X0_MISC_GCR - A3X0_MISC_MODULE_BASE +
				A3X0_MISC_MODULE_BASE_PA,
				0x02, LED_DEVICE))
		return -EBUSY;
	if(misc_register(&led_dev)){
		return -ENODEV;
	}

	PRINT_INFO(VERSION "\n");

	return 0;
}

static void __exit 
led_exit(void)
{
	DEBUG_FUNC();
	misc_deregister(&led_dev);
	release_mem_region(A3X0_MISC_GCR
			   - A3X0_MISC_MODULE_BASE + A3X0_MISC_MODULE_BASE_PA,
			   0x02);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(led_init);
module_exit(led_exit);
